#!/usr/bin/env python
from mpi4py import MPI
import numpy as np
import time
import sys
from common import *
import logging


if '-v' in sys.argv:
    logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)
else:
    logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)


comm = MPI.COMM_WORLD
worldSize = comm.Get_size()
rank = comm.Get_rank()
processorName = MPI.Get_processor_name()
TaskMaster = 0


def EP_encode_B(M, P1, P2, X):
    S = M.shape
    f = np.zeros((S[0] // P1, S[1] // P2))
    for x2 in range(1, P2 + 1):
        for x1 in range(1, P1 + 1):
            f += M[((P1 - x1) * S[0] // P1): ((P1 + 1 - x1) * S[0] // P1), ((x2 - 1) * S[1] // P2): (x2 * S[1] // P2)] * (float(X) ** (P1 * x2 + x1))
    return f


def EP_encode_A(M, P, X):
    S = M.shape
    f = np.zeros((S[0] // P[0], S[1] // P[1]))
    for x0 in range(1, P[0] + 1):
        for x1 in range(1, P[1] + 1):
            f += M[((x0 - 1) * S[0] // P[0]): (x0 * S[0] // P[0]), ((x1 - 1) * S[1] // P[1]): (x1 * S[1] // P[1])] * (float(X) ** (P[1] * P[2] * x0 + x1))
    return f


def EP_encode_A1_Bm(M, P, X):
    f = []
    f.append(EP_encode_A(M[0], P[:3], X))
    for i in range(1, len(M)):
        f.append(EP_encode_B(M[i], P[1], P[2], X))
    return f

def traditional_decode(R, P, X):
    m = 2
    threshold = np.prod(P) + np.prod(P[1:-1]) - 1
    M = np.ones((threshold, threshold))
    power = np.prod(P[1 : m + 1]) + np.prod(P[1 : m])
    base = power
    power += np.prod(P[2: m]) * 2
    base += np.prod(P[2 : m]) * (P[1] + 1)
    base = base - power
    for i in range(0, threshold):
        for j in range(0, threshold):
            M[i, j] = float(X[i]) ** (j + power)
    rows = []
    for x0 in range(P[0]):
        for xm in range(P[m]):
            rows.append(base + np.prod(P[1: m + 1]) * x0 + np.prod(P[1: m]) * xm)
    rows = [int(x) for x in rows]
    R = np.reshape(R, (threshold, -1))
    M = np.linalg.pinv(M)
    M = M[rows, :]
    return np.dot(M, R)


m = int(sys.argv[1])
size = int(sys.argv[2])
P0 = int(sys.argv[3])
P1 = int(sys.argv[4])
P2 = int(sys.argv[5])
P = [P0, P1, P2]

if rank == TaskMaster:
    M = []
    Original_result = np.ones((size, size))
    for i in range(m):
        M.append(np.ones((size, size)))
    for i in range(m-2):
        Original_result = np.dot(Original_result,np.ones((size, size)))

m_times = 1
c = 35.0

logging.debug("[%d] starts" % (rank,))
while(1):
    if(m_times < m):
        R = []
        P_mtimes = []
        for i in range(0, 3):
            P_mtimes.append(P[i])
        threshold = P_mtimes[0] * P_mtimes[1] * P_mtimes[2] + P_mtimes[1] - 1

        for i in range(worldSize + 1):
            R.insert(i, "")

        if rank != TaskMaster:
            logging.debug("[%d] start round %d" % (rank, m_times))
            if m_times == 1:
                f = comm.recv(source=TaskMaster, tag=100)
                logging.debug("[%d] barrier" % (rank,))
                comm.barrier()
                R.insert(rank - 1, np.dot(f[0], f[m_times]))
            else:
                f0 = comm.recv(source=TaskMaster, tag=100)
                R.insert(rank - 1, np.dot(f0, f[m_times]))
            del R[rank]
            logging.debug("[%d] sending result to master" % (rank,))
            R[rank - 1].reshape(1, -1)
            request = comm.isend(R[rank - 1], dest=TaskMaster, tag=m_times)
            logging.debug("[%d] finish round %d" % (rank, m_times))
            m_times += 1
            if(m_times == m):
                exit(0)

        if rank == TaskMaster:
            logging.debug("[%d] round %d starts" % (rank, m_times))
            Result = []
            index = []
            M_mtimes = []
            for i in range(m_times - 1, m_times + 1):
                M_mtimes.append(M[i])
            S = []
            for i in range(len(M_mtimes)):
                S.append(M_mtimes[i].shape)
            for i in range(1, worldSize):
                if m_times == 1:
                    f = EP_encode_A1_Bm(M, P, i / c)
                    comm.send(f, dest=i, tag=100)
                else:
                    f = EP_encode_A(M[m_times - 1], P_mtimes, i / c)
                    comm.send(f, dest=i, tag=100)
            if m_times == 1:
                logging.debug("[%d] barrier" % (rank,))
                comm.barrier()
                start = time.time()
            while (len(index) < threshold):
                st = MPI.Status()
                comm.Probe(tag=m_times, status=st)
                i = st.source
                tag = st.tag
                R.insert(i - 1, comm.recv(source=i, tag=m_times))
                logging.debug("[%d] receiving result from %d" % (rank, i))
                del R[i]
                index.append(i / c)
            logging.debug("[%d] round %d ends" % (rank, m_times))
            index.sort()
            for j in index:
                Result.append(R[int(j * c) - 1])
            m_times += 1
            if (m_times == m):
                traditional_decode(Result, P_mtimes, index)
                #print_time(start)
                exit(0)
            else:
                RESULT_m_times = traditional_decode(Result, P_mtimes, index)
                R1, R2 = RESULT_m_times.shape
                M[m_times - 1] = np.reshape(RESULT_m_times, (size, int(R1 * R2 / size)))
                print_time(start)
                if m_times == m - 1:
                    different = np.sum(np.absolute(Original_result - M[m_times - 1]))
                    numerical_stability = different / np.sum(Original_result)
                    #print(M[m_times - 1])
                    #print(Original_result)
                    #print('Numerical stability = ' + str(numerical_stability))
                
                    