#!/usr/bin/env python
from mpi4py import MPI
import math
import numpy as np
import time
import sys
from common import *


comm = MPI.COMM_WORLD
worldSize = comm.Get_size()
rank = comm.Get_rank()
processorName = MPI.Get_processor_name()
TaskMaster = 0


def chain_code(M, P, X):
    m = len(M)
    S = []
    for i in range(m):
        S.append(M[i].shape)
    if m == 2:
        f = []
        for i in range(m):
            # TODO: check if values in P are legal
            f.append(np.zeros((S[i][0] // P[i], S[i][1] // P[i + 1])))
        for x0 in range(1, P[0] + 1):
            for x1 in range(1, P[1] + 1):
                f[0] += M[0][((x0 - 1) * S[0][0] // P[0]): (x0 * S[0][0] // P[0]),
                        ((x1 - 1) * S[0][1] // P[1]): (x1 * S[0][1] // P[1])] * (float(X) ** (P[1] * P[2] * x0 + x1))
        for x2 in range(1, P[2] + 1):
            for x1 in range(1, P[1] + 1):
                f[1] += M[1][((P[1] - x1) * S[1][0] // P[1]): ((P[1] + 1 - x1) * S[1][0] // P[1]),
                        ((x2 - 1) * S[1][1] // P[2]): (x2 * S[1][1] // P[2])] * (float(X) ** (P[1] * x2 + x1))
        return f
    elif m % 2 == 1:
        f = chain_code(M[0:-1], P[0:-1], X)
        # print(f)
        f.append(np.zeros((S[m - 1][0] // P[m - 1], S[m - 1][1] // P[m])))
        for xm in range(1, P[m] + 1):
            for xm1 in range(1, P[m - 1] + 1):
                f[-1] += M[m - 1][
                         ((P[m - 1] - xm1) * S[m - 1][0] // P[m - 1]):((P[m - 1] + 1 - xm1) * S[m - 1][0] // P[m - 1]),
                         ((xm - 1) * S[m - 1][1] // P[m]): (xm * S[m - 1][1] // P[m])] * (
                                     float(X) ** (np.prod(P[0:m]) * xm + np.prod(P[1:m - 1]) * xm1))
        return f
    elif m % 2 == 0:
        f = chain_code(M[1:], P[1:], X)
        f = [np.zeros((S[0][0] // P[0], S[0][1] // P[1]))] + f
        for x0 in range(1, P[0] + 1):
            for x1 in range(1, P[1] + 1):
                f[0] += M[0][((x0 - 1) * S[0][0] // P[0]): (x0 * S[0][0] // P[0]),
                        ((P[1] - x1) * S[0][1] // P[1]): ((P[1] + 1 - x1) * S[0][1] // P[1])] * (
                                    float(X) ** (np.prod(P[1:m + 1]) * x0 + np.prod(P[2:m]) * x1))
        return f
    else:
        raise Exception("invalid m (m = %s)" % m)


def chain_decode(R, P, X):
    m = len(P) - 1
    threshold = np.prod(P) + np.prod(P[1:-1]) - 1
    M = np.ones((threshold, threshold))
    if m % 2 == 0:
        power = np.prod(P[1: m + 1]) + np.prod(P[1: m])
        base = power
        power += np.prod(P[2: m]) * 2
        base += np.prod(P[2: m]) * (P[1] + 1)
        for l in range(0, int(math.ceil(m / 2.0 - 2 + 1))):  # problem
            power += np.prod(P[2 + l: m - 2 - l + 1]) * 2 + np.prod(P[3 + l: m - 2 - l + 1]) * 2
            base += np.prod(P[2 + l: m - 2 - l + 1]) * (P[m - 1 - l] + 1) + np.prod(P[3 + l: m - 2 - l + 1]) * (
                        P[2 + l] + 1)
    else:
        power = np.prod(P[0: m]) + np.prod(P[1: m])
        base = power
        for l in range(0, int(math.ceil(m / 2.0 - 1.5 + 1))):
            power += np.prod(P[1 + l: m - 2 - l + 1]) * 2 + np.prod(P[2 + l: m - 2 - l + 1]) * 2
            base += np.prod(P[1 + l: m - 2 - l + 1]) * (P[m - 1 - l] + 1) + np.prod(P[2 + l: m - 2 - l + 1]) * (
                        P[1 + l] + 1)
    base = base - power
    for i in range(0, threshold):
        for j in range(0, threshold):
            M[i, j] = float(X[i]) ** (j + power)
    rows = []
    for x0 in range(P[0]):
        for xm in range(P[m]):
            if m % 2 == 0:
                rows.append(base + np.prod(P[1: m + 1]) * x0 + np.prod(P[1: m]) * xm)
            else:
                rows.append(base + np.prod(P[0: m]) * xm + np.prod(P[1: m]) * x0)
    rows = [int(x) for x in rows]
    R = np.reshape(R, (threshold, -1))
    M = np.linalg.pinv(M)
    M = M[rows, :]
    return np.dot(M, R)


m = int(sys.argv[1])
size = int(sys.argv[2])
partition = int(sys.argv[3])

if rank == TaskMaster:
    M = []
    #for i in range(m):
    A = np.ones((size, size))
    AT = np.transpose(A)
    X = np.ones((size, 1))
    M.append(AT)
    M.append(A)
    #M.append(X)

P = [partition] * (m + 1)
#P = [2, 2, 2]

R = []
threshold = np.prod(P) + np.prod(P[1:-1]) - 1

for i in range(worldSize + 1):
    R.insert(i,"")
#print(m)
c = 35.0

if rank != TaskMaster:
    f = comm.recv(source=TaskMaster, tag=11)
    comm.barrier()
    R.insert(rank - 1, matrix_prod(f))
    del R[rank]
    request = comm.send(R[rank - 1], dest=TaskMaster, tag=0)
    exit(0)
if rank == TaskMaster:
    print(threshold)
    Result = []
    index = []
    for i in range(1, worldSize):
        f = chain_code(M, P, i / c)
        #print(str(f) + '=' + str(i))
        comm.send(f, dest=i, tag=11)
    comm.barrier()
    start = time.time()
    while(len(index) < threshold):
        st = MPI.Status()
        comm.Probe(status = st)
        i = st.source
        R.insert(i - 1, comm.recv(source=i, tag=0))
        del R[i]
        index.append((i) / c)
    index.sort()
    for j in index:
        #print(j* c)
        Result.append(R[int(j * c) - 1])
    # print("result:")
    # print(chain_decode(Result, P, index))
    #print(chain_decode(Result, P, index))

    # print(matrix_prod(M))
    print_time(start)
    exit(0)
